'''


'''
import logging
import os
import shutil
from copy import deepcopy as dcopy

import numpy as np

import torch
import torch.optim as optim
import torch.utils.data as D

import tools # since T is used as a variable in this module, refrain from "import tools as T"
import tools.torch
import tools.modules

from . import utils as U
from . import evaluation as E
from . import model as M

# __all__ = [
# 'Trainer'
# ]

# %%
log = logging.getLogger(__name__)

# %%
def validate_cfg_val(cfg_val):
    cfg_val = dcopy(dict(cfg_val)) if cfg_val is not None else {}
    if 'forward_f' not in cfg_val:
        cfg_val['forward_f'] = None
    return cfg_val

class Trainer:
    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir='network', cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
        """
        - train network with early stopping, given hyperparameters
        - save/load network with necessary preprocessing pipelin

        Parameters
        ----------
        network : torch.nn.Module object
            Network to train

        train_dataset: torch.utils.data.Dataset object
            This dataset is used for training.
            Recommended return format from dataset is dictionary form, although tuples or raw tensors are supported.
            Example: {'x': x, 'y': y}
            
            train_dataset is fed into torch.utils.data.DataLoader object, where it is batched then passed to self.forward(), with the following pattern depending on train_dataset return type:
                torch.Tensor: self.forward(data)
                tuple or list: self.forward(*data)
                dict: self.forward(**data)

        val_dataset: torch.utils.data.Dataset object or dict of such objects
            Dataset used for validation (Early stopping)
            forward mechanism from val_dataset may be different from train_dataset, which could be defined in cfg_val['forward_f']        

        cfg_train: dict-like object with following keys:
            lr: float
            batch_size: int
            weight_decay: bool (optional)
            cv_step: int (optional, recommended when val_dataset is given)
        
        cfg_val: (required if val_dataset is given) dict-like object with following keys:
            increase_better: Bool
            patience: int
            batch_size: int

            forward_f: callable, forward pass that might be different from training forward pass (optional)
            target_dataset: str (optional, if val_dataset is dict for multiple datasets)
            criterion: str (optional, if val_metrics is dict for multiple metrics) 
        
        criterion: nn.Module object
            Defines the loss function.
            If parameters exist within criterion, it is moved to the corresponding device where network lies.
        
        network_dir: str or Pathlike object, default='network'
            The directory to save the nn.
            If early stopping is used, a subdirectory will be created that keeps the models during training.

        Examples
        --------
        # Recommended way of making hs.Trainer object is like the following:

        kwargs = {'network': Network(), 'train_dataset': train_dataset, 'validation': None, 'cfg_train': cfg.train,
                'criterion': criterion, 'network_dir': path['network'], 'verbose': True, 'amp': True}
        trainer = hs.Trainer(**kwargs)

        # To make a custom Trainer for different neural network applications, just define the forward function:

        class MyTrainer(hs.Trainer):
            def forward(self, x, y, z):
                w = self.network(x,y)
                loss = self.criterion(w, z)
                return loss
        """

        # Store configurations
        self.network = network
        self.train_dataset = train_dataset
        self.cfg_train = dcopy(dict(cfg_train))
        self.criterion = criterion
        self.network_dir = network_dir

        self.cfg_val = validate_cfg_val(cfg_val)
        self.val_dataset = val_dataset
        self.val_metrics = val_metrics
        self.earlystopper = EarlyStopper(increase_better=cfg_val['increase_better'], patience=cfg_val['patience'], save_dir=os.path.join(self.network_dir, 'network_temp')) if self.val_dataset is not None else None
        if 'target_dataset' in self.cfg_val:
            if type(self.val_dataset)==dict:
                assert self.cfg_val['target_dataset'] in self.val_dataset, f'[Validation setup] target_dataset: {self.cfg_val["target_dataset"]} not found in val_dataset: {self.val_dataset.keys()}'
            assert ~((self.val_dataset is not None) ^ (self.val_metrics is not None)), f'val_dataset and val_metrics must be given together or neither. val_dataset: {self.val_dataset}, val_metrics: {self.val_metrics}'

        self.name = name
        self.verbose = verbose
        self.amp = amp
        self.reproduce = reproduce

        # Initializations
        if self.network_dir is not None:
            if os.path.exists(self.network_dir): log.warning(f'path: {self.network_dir} exists. Be careful')
            os.makedirs(self.network_dir, exist_ok=True)

        self.train_meter = tools.modules.AverageMeter() # Tracks loss per epoch
        self.loss_tracker = tools.modules.ValueTracker() # Tracks loss over all training
        self.set_model()
        self.reset()

        self._targets_type = self.infer_targets_type()

    def reset(self):
        self.iter = 0
        self.n_batch=0

    def print(self, content):
        if self.verbose:
            print(content)
        else:
            log.info(content)

    def set_model(self):
        '''Set miscellaneous variables'''
        self.model = M.BaseModel(self.network, amp=self.amp)
        self.misc = tools.TDict()
        if self.train_dataset is not None:
            if hasattr(self.train_dataset, 'get_f'): # get_f equal to preprocessing pipeline before feeding to network
                self.print('get_f found in train_dataset')
                self.misc.get_f = self.train_dataset.get_f

    def validate(self):
        patience_end = False
        self.print(f'[Node: {self.name}] Validation')
        if self.val_dataset is not None:
            if type(self.val_dataset)==dict:
                scores = {}
                for name, dataset in self.val_dataset.items():
                    results = E.forward_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], forward_f=self.cfg_val['forward_f'], amp=self.amp)
                    isnan = any([np.isnan(v).any() for v in results.values()]) if isinstance(results, dict) else np.isnan(np.concatenate(results)).any()
                    if isnan:
                        log.warning(f'[Node: {self.name}][validate] NaN found in results. Skipping validation...')
                        score = {self.cfg_val['criterion']: self.earlystopper.best_score}
                    else:
                        score = E.evaluate(results, self.val_metrics)
                    scores[name] = score
                    self.print(f'[dataset_name: {name}] Validation Score: {score}')
                score = scores[self.cfg_val['target_dataset']]
            else:
                results = E.forward_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], forward_f=self.cfg_val['forward_f'], amp=self.amp)
                isnan = any([np.isnan(v).any() for v in results.values()]) if isinstance(results, dict) else np.isnan(np.concatenate(results)).any()
                if isnan:
                    log.warning(f'[Node: {self.name}][validate] NaN found in results. Skipping validation...')
                    score = {self.cfg_val['criterion']: self.earlystopper.best_score}
                else:
                    score = E.evaluate(results, self.val_metrics)
                self.print(f'Validation Score: {score}')
            
            if self.earlystopper is not None:
                if type(score) is dict:
                    score_ = score[self.cfg_val['criterion']]
                else:
                    score_ = score
                    
                patience_end = self.earlystopper.step(self, score_)
                if patience_end:
                    self.print('patience met, flushing earlystopper history')
                    self.earlystopper.history.reset()
                    
        torch.cuda.empty_cache()
        self.network.train()
        return patience_end

    def generate_loader(self):
        if self.train_dataset is not None:
            reproduce_kwargs = U.reproducible_worker_dict() if self.reproduce else {}
            drop_last = len(self.train_dataset) % self.cfg_train['batch_size'] == 1 # To avoid batch normalization layers from raising exceptions
            self.loader = D.DataLoader(self.train_dataset, batch_size=self.cfg_train['batch_size'], shuffle=True, drop_last=drop_last, **reproduce_kwargs)
        else:
            log.warning('No train_dataset found. Skipping generate_loader()')

    def train(self, epoch=None, new_op=True, no_val=False, step=None, reset_loss_tracker=False, device=None):
        """
        Trains the network with the specified duration.

        Parameters
        ----------
        epoch : int, default=None
            Number of epochs to train the network.
            If None, the value is taken from self.cfg_train['epoch']
            Must be mutually exclusive with "step" argument.

        new_op : bool, default=True
            Whether to create a new optimizer.
            If False, the optimizer is not initialized, keeping the previous state.

        no_val : bool, default=False
            Whether to skip validation during training.
            If True, the validation is skipped even when val_dataset is given.
        
        step : int, default=None
            Number of steps to train the network.
            Must be mutually exclusive with "epoch" argument.
        
        reset_loss_tracker : bool, default=False
            Whether to reset the loss_tracker before training.

        device : torch.device, default=None
            Explicity Device to train the network, recommended to default to None. 
            If None, the device is taken from the network.

        Returns
        -------
        loss_tracker : tools.modules.ValueTracker

        """
        assert epoch is None or step is None, f'only one of epoch or step can be specified. received epoch: {epoch}, step: {step}'
        if step is None:
            horizon = 'epoch'
            if epoch is None: # When neither epoch or step are specified
                assert 'epoch' in self.cfg_train, 'key "epoch" must be provided in cfg_train, or the argument "epoch" must be provided \
                                                when argument "step" is not specified.'
                epoch = self.cfg_train['epoch']
            log.debug(f'[Node: {self.name}] train for {epoch} epochs')
        else:
            assert epoch is None, f'Either epoch or step must be specified. Received epoch: {epoch}, step: {step}'
            horizon = 'step'
            step = step
            log.debug(f'[Node: {self.name}] train for {step} steps')

        self._device = device if device is not None else tools.torch.get_device(self.network)
        self.criterion = self.criterion.to(self._device)
        if reset_loss_tracker: self.loss_tracker.reset()
        self.generate_loader()
        '''
        There may be one or more loaders, but self.loader is the standard of synchronization with batches.
        Either return multiple values from dataset, or modify self.forward to use other loaders.
        '''

        if self.val_dataset is not None:
            if 'cv_step' not in self.cfg_train:
                self.print('val_dataset is given but "cv_step" not specified in cfg_train. Defaults to 1 epoch')
                self.cfg_train['cv_step'] = len(self.loader)
            self.validate() # initial testing

        # Make new optimizer
        if new_op or not hasattr(self, 'op'):
            if not hasattr(self, 'op') and not new_op:
                log.warning("new_op=False when there's no pre-existing optimizer. Ignoring new_op...")
            
            l_kwargs = ['lr', 'weight_decay', 'amsgrad']
            kwargs = {key: self.cfg_train[key] for key in l_kwargs if key in self.cfg_train}
            self.op = optim.Adam(self.network.parameters(), **kwargs)

        if horizon == 'epoch':
            self._update_epoch(T=epoch, no_val=no_val)
        elif horizon=='step':
            self._update_step(T=step, no_val=no_val)

        # TODO: Return criterion back to its original device, meaning we have to store its previous device info
        self.criterion = self.criterion.cpu()
        return self.loss_tracker

    def _update_epoch(self, T, no_val=False):

        _iter = 0
        for epoch in range(1, T+1):
            self.train_meter.reset()
            self.epoch_f()
            for batch_i, data in enumerate(self.loader, 1):
                self.iter += 1
                _iter += 1
                loss = self.update(data)

                self.print(f'[Node: {self.name}][iter_sum: {self.iter}][Epoch: {epoch}/{T}][Batch: {batch_i}/{len(self.loader)}][Loss: {loss:.7f} (Avg: {self.train_meter.avg:.7f})]')
                self.loss_tracker.step(self.iter, loss)

                # Validation
                if (self.val_dataset is not None) and (not no_val) and (_iter % self.cfg_train['cv_step']==0):
                    patience_end = self.validate()
                    if patience_end: # If patience has reached, stop training
                        self.print('Patience met, stopping training')
                        return None # return None since double break is impossible in python

    def _update_step(self, T, no_val=False):
        if hasattr(self, '_loader_inst'): del self._loader_inst # Load data from the 1st batch, start fresh

        for _iter in range(1, T+1):
            # Get Data
            try:
                if hasattr(self, '_loader_inst'):
                    data = next(self._loader_inst)
                    self.n_batch += 1
                else:
                    raise StopIteration
            except StopIteration as e:
                self.train_meter.reset()
                self.epoch_f()
                self._loader_inst = iter(self.loader)
                data = next(self._loader_inst)
                self.n_batch = 1

            self.iter += 1
            loss = self.update(data)

            self.print(f'[iter_sum: {self.iter}][Iter: {_iter}/{T}][Batch: {self.n_batch}/{len(self.loader)}][Loss: {loss:.6f} (Avg: {self.train_meter.avg:.6f})]')
            self.loss_tracker.step(self.iter, loss)

            # Validation
            if (self.val_dataset is not None) and (not no_val) and (_iter % self.cfg_train['cv_step']==0):
            # if (not no_val) and (self.iter % self.cfg_train.cv_step==0):
                patience_end = self.validate()
                if patience_end: # If patience has reached, stop training
                    self.print('Patience met, stopping training')
                    return None # return None since double break is impossible in python

    def update(self, data):
        """
        - Perform single update (forward/backward pass + gradient descent step) with the given data.
        - Store loss in self.train_meter
        - This is where a lot of errors happen, so there's a pdb to save time.
          When there's error, use pdb to figure out the shape, device, tensor dtype, and more.

        Parameters
        ----------
        data : tuple, list, or dict of tensors (Batch of data)
            This is received from a torch.utils.Data.DataLoader
            Depending on the given format, the data is fed to the forwad pass

        Returns
        -------
        loss : float
        """
        # Automatic Mixed Precision (AMP, float32 -> float16) for acceleration
        if self.amp:
            with torch.autocast(device_type=self._device.type):
                loss, N = self._forward(data)
        else:
            loss, N = self._forward(data)

        self.op.zero_grad()
        loss.backward()
        self.op.step()

        loss = loss.item()
        self.train_meter.step(loss, N)
        return loss

    def _forward(self, data): # TODO: change to function decorator?
        '''
        Pseudo function to:
        1. Move data to self._device
        2. Support passing tuple or dict from a batch from dataloader to forward()
        '''
        datatype = type(data)
        # When data is given as a tuple/list
        if datatype is tuple or datatype is list:
            data = [x.to(self._device) for x in data]
            N = len(data[0]) # number of data in batch
            loss = self.forward(*data)

        # When data is given as a dict
        elif datatype is dict:
            data = {key: value.to(self._device) for key, value in data.items()}
            N = len(next(iter(data.values()))) # number of data in batch
            loss = self.forward(**data)

        elif datatype is torch.Tensor:
            data = data.to(self._device)
            N = len(data)
            loss = self.forward(data)
            
        else:
            raise Exception(f'return type from dataset must be one of [torch.Tensor, tuple, list, dict], received: {datatype}')
        return loss, N

    def forward(self, x, y):
        """
        Forward pass: Receive data and pass values to criterion.
        Define forward for custom trainers.

        May return tuple or dictionary, whichever will be feeded to criterion.

        Defaults to args[0] feeding to self.network (assuming x), and args[1] feeding to self.criterion (assuming y)
        """
        
        y_hat = self.network(x)
        loss = self.criterion(y_hat, y)

        return loss

    def epoch_f(self):
        # TODO: rename to forward hook? epoch_hook?
        '''function to call every other epoch. May be used in child class'''
        pass

    def infer_targets_type(self):
        if self.train_dataset is not None and hasattr(self.train_dataset, 'targets_type'):
            return self.train_dataset.targets_type
        elif self.val_dataset is not None:
            if type(self.val_dataset)==dict:
                for name, dataset in self.val_dataset.items():
                    if hasattr(dataset, 'targets_type'):
                        return dataset.targets_type
            elif hasattr(self.val_dataset, 'targets_type'):
                return self.val_dataset.targets_type
        else:
            return None

    def post_train(self, val_dataset=None):
        if val_dataset is None:
            assert self.val_dataset is not None, f'val_dataset must be given if self.val_dataset is None'
            if type(self.val_dataset)==dict:
                val_dataset = self.val_dataset[self.cfg_val['target_dataset']]
            else:
                val_datset = self.val_dataset

        if self._targets_type == 'binary':
            self.print(f'[Node: {self.name}][post_train] binary classfication: choose threshold based on validation set')
            self.threshold = E.binary_threshold(self, val_dataset)
        else:
            self.print(f'[Node: {self.name}][post_train] no post-train procedure for _targets_type: {self._targets_type}')

    '''
    Save/Load functions
    state_dict()
    load_state_dict()

    save()
    save_best()
    load()
    '''
    def state_dict(self):
        state_dict = {
        'loss_tracker': self.loss_tracker,
        }
        if 'get_f' in self.misc:
            state_dict['misc.get_f'] = self.misc.get_f
        if hasattr(self, 'threshold'):
            state_dict['threshold'] = self.threshold
        if self.earlystopper is not None:
            state_dict['earlystopper.history'] = self.earlystopper.history
        return state_dict

    def load_state_dict(self, state_dict):
        state_dict = dcopy(state_dict)

        if 'misc.get_f' in state_dict.keys():
            self.print(f'Updating get_f: {state_dict["misc.get_f"]}')
            self.misc.get_f = state_dict['misc.get_f']
            if self.train_dataset is not None:
                self.print('updating get_f to loader...')
                self.train_dataset.get_f = self.misc.get_f
            del state_dict['misc.get_f']

        if 'threshold' in state_dict.keys():
            self.print(f'Updating threshold: {state_dict["threshold"]}')
            self.threshold = state_dict['threshold']
            del state_dict['threshold']

        self.__dict__.update(state_dict) # update attributes
        
    def load_best_model(self):
        if self.earlystopper is not None:
            if self.earlystopper.best_network != None:
                self.print(f'Loading best network: {self.earlystopper.best_network}')
                self.network.load_state_dict(torch.load(self.earlystopper.best_network))
            else:
                self.print(f'No best_network found in self.earlystopper. No network loaded.')
        else:
            self.print('self.earlystopper not defined. No network loaded.')

    def save(self, path=None, best=True):
        '''
        Save the following:
        state_dict() -> path/state_dict.p
        network.state_dict() -> path/network.pt
        '''
        path = self.network_dir if path is None else path
        self.print(f'[Node: {self.name}]')

        state_dict_path = os.path.join(path, 'model.p')
        state_dict = self.state_dict()
        tools.save_pickle(state_dict, state_dict_path)
        self.print(f'[save] Saving Model info to: {state_dict_path}')
        self.print(f'[save] state_dict: {list(state_dict.keys())}')

        network_path = os.path.join(path, 'network.pt')
        if best:
            if self.earlystopper is None:
                self.print(f'[save][best: {best}] earlystopper not defined. Saving current network -> [{network_path}]')
                torch.save(self.network.state_dict(), network_path)
            else:
                if self.earlystopper.best_network != None:
                    self.print(f'[save][best: {best}] Saving [{self.earlystopper.best_network}] -> [{network_path}]')
                    shutil.copy(self.earlystopper.best_network, network_path)
                else:
                    self.print(f'[save][best: {best}] no best_network in self.earlystopper, Saving current network -> [{network_path}]')
                    torch.save(self.network.state_dict(), network_path)
        else:
            self.print(f'[save][best: {best}] Saving network to: {network_path}')
            torch.save(self.network.state_dict(), network_path)

    def load(self, path=None):
        '''
        load state_dict() and network:
        path/node.p -> state_dict
        path/network.pt -> network.state_dict
        '''
        path = self.network_dir if path is None else path
        self.print(f'[Node: {self.name}]')

        state_dict_path = os.path.join(path, 'node.p')
        self.print(f'[load] Loading from path: {state_dict_path}')
        state_dict = tools.load_pickle(state_dict_path)
        self.print(f'[load] Updating: {list(state_dict.keys())}')
        self.load_state_dict(state_dict)

        network_path = os.path.join(path, 'network.pt')
        self.print(f'[load] Loading network from: {network_path}')
        self.network.load_state_dict(torch.load(network_path))

class EarlyStopper():
    """
    Returns True if the metric has stopped improving for the given number of patience steps.
    Saves network in the given path and remembers which network is the best.

    Parameters
    ----------
    arg : array-like of shape (n_samples,), default=None
        Argument explanation.
        If ``None`` is given, those that appear at least once
        .. versionadded:: 0.18
    
    Functions
    -------
    step()
        returns True/False
    """

    def __repr__(self):
        return f'<EarlyStopper>\npatience: {self.patience}\nincrease_better: {self.increase_better}'

    def __init__(self, increase_better, patience, save_dir='network_temp', discard_best=True):
        self.increase_better = increase_better
        self.patience = patience
        self.save_dir = save_dir
        self.discard_best = discard_best

        os.makedirs(self.save_dir, exist_ok=True)
        self.history = tools.modules.ValueTracker() # For now, only track single score
        self.reset()

    def step(self, trainer, score, filename=None):
        filename = f'network_{trainer.iter}.pt' if filename is None else filename

        self.history.step(trainer.iter, score)

        # Save best network
        if isbetter(score_best=self.best_score, score=score, increase_better=self.increase_better):
            path = os.path.join(self.save_dir, filename)

            log.info(f'[EarlyStopping] New best score: {self.best_score} -> {score}')
            log.info(f'[EarlyStopping] Saved network: {path}')
            torch.save(trainer.network.state_dict(), path)
            if self.discard_best and self.best_network!=None:
                os.remove(self.best_network)

            self.best_score = score
            self.best_network = path

            # Clear patience tracking
            self.history.reset()

        # patience_end == True when best network did not appear for self.patience times
        patience_end = self.patience == len(self.history)
        log.info(f'[EarlyStopping][patience: {len(self.history)}/{self.patience}]')

        return patience_end

    def reset(self):
        # log.info('[EarlyStopper] Resetting...')
        self.history.reset()
        self.best_network = None
        if self.increase_better:
            self.best_score = -float('inf')
        else:
            self.best_score = float('inf')

def isbetter(score_best: float, score: float, increase_better: bool):
    return (not increase_better and score<score_best) or (increase_better and score>score_best)



# %%
import logging

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as D

import tools as T

# TODO: Remove everything except feature_visualize() (optimize(), _optimize(), RandomSampler, ...)

# %%
log = logging.getLogger(__name__)

# %%
def feature_visualize(objective, data, maximize=True, lr=1e-1, updates=100, threshold=None, preprocess=None, regularization=None, verbose=False):
    """
    Feature visualization give objective

    Parameters
    ----------
    objective : nn.Module that takes data as input and outputs a single value to optimize.
        the parameters of objective follows the device of data

    data: torch.Tensor of shape (n_sample, *input_shape)
        Could be sample data or noise
    
    maximize: bool, default=True
        If True, then output of objective is multiplied my -1 (since loss is minimized in pytorch)

    lr: float, default=1e-1
        Learning rate

    updates: int, default=100
        Number of updates

    threshold: float, default=None
        If given, then optimization stops when output of objective is larger than threshold (for maximize=True), or less than threshold (maximize=False).
    
    preprocess: callable, default=None
        Preprocess the input data before passing to the objective function.
        The feature visualization will return the preprocessed data.

    regularization: callable, default=None
        Regularize the input data before passing to the objective function.
        
    Returns
    -------
    data : torch.Tensor
    """

    data = data.clone() # Original tensor is not modified
    data.requires_grad = True

    objective.to(data.device)
    objective.eval()

    # Gradient descent not ascent, so negate the objective if maximize=True
    if maximize: threshold = -threshold if threshold is not None else None

    # Thresholded flag for batch data when objective threshold is given
    if threshold is not None:
        thresholded = torch.zeros(len(data), dtype=torch.bool, device=data.device)
        data_optimized = torch.empty_like(data)
    
    # Default is doing nothing
    preprocess = nn.Identity() if preprocess is None else preprocess
    regularization = nn.Identity() if regularization is None else regularization

    # Defaults to Adam optimizer
    optimizer = optim.Adam([data], lr=lr) 

    l_history = []
    for i in range(1,updates+1):

        # Forward pass
        data_p = preprocess(data)
        data_r = regularization(data_p)
        loss = objective(data_r)

        if maximize:
            loss = -loss

        # Check loss < threshold to stop SGD update
        l_history.append(data_p.detach().clone())
        if threshold is not None:
            thresholded_ = (loss.detach() < threshold)
            store_idx = thresholded_ & ~thresholded
            data_optimized[store_idx] = data[store_idx].detach().clone()
            thresholded[store_idx] = True
            
            if thresholded.all(): break

        # SGD Update
        loss = loss.mean()
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
        if verbose:
            loss = -loss.item() if maximize else loss.item()
            print(f'[{i}/{updates}][Loss: {loss}]')

    if threshold is not None:
        data_optimized[~thresholded] = data[~thresholded].detach().clone()
        data = data_optimized
    
    # Pass preprocess
    with torch.no_grad():
        features = preprocess(data)

    return features

class Output_i(nn.Module):
    """
    Output objective

    Parameters
    ----------
    model : nn.Module object
        Model to optimize.
    class_i: int
        Output class index.
    maximize: bool, default=True
        If True, then forward returns negated value (since loss is minimized in pytorch),
    suppress: bool, default=False
        If True, then forward adds non-negated values of other classes.
        Suppresses other class activations.

    Examples
    --------
    >>> log.info('hello')
    hello
    """
    def __init__(self, model, class_i, maximize=True, suppress=False):
        super().__init__()
        self.model = model
        self.class_i = class_i
        self.maximize = maximize
        self.suppress = suppress

    def forward(self, x):
        y_hat = self.model(x)
        objective = y_hat[:, self.class_i]
        if self.suppress:
            class_i_suppress = list(range(y_hat.shape[-1]))
            class_i_suppress.remove(self.class_i)
            objective = objective - y_hat[:, class_i_suppress].mean(-1)

        if self.maximize:
            objective = -objective
            
        return objective

# DEPRECATED: Everything can be done with much cleaner code of feature_visualize()

# class RandomSampler():
#     '''
#     Sample random data
#     '''
#     allowed_methods = ['rand', 'randn']
#     def __init__(self, input_shape, method='rand'):
#         self.input_shape = input_shape
#         self.method = method
#         assert method in self.allowed_methods, f'allowed methods are: {self.allowed_methods}\nreceived: {method}'

#     def sample(self, n_sample, method=None):
#         method = self.method if method is None else method
#         if method == 'rand':
#             sample = torch.rand(n_sample, *self.input_shape)
#         elif method == 'randn':
#             sample = torch.randn(n_sample, *self.input_shape)
#         return sample

# # %%
# def _optimize(objective, sample, updates, Optimizer, optimizer_kwargs, preprocess=None, regularization=None, threshold=None, return_best=False, verbose=False):
#     device_original = sample.device
#     device = T.torch.get_device(objective)
#     sample = sample.clone().detach().to(device) # clone sample so that original tensor will not be modified.
#     sample.requires_grad_(True)

#     # if not sample.requires_grad:
#     #     warnings.warn('Provide sample with sample.requires_grad_(True).\
#     #     _optmize() function will automatically setting requires_grad to True, but this may resuilt in undesired results.')

#     optimizer_ = Optimizer([sample], **optimizer_kwargs)

#     if preprocess is not None:
#         log.info('preprocess specified')
#     if regularization is not None:
#         log.info('regularization specified')

#     if verbose:
#         with torch.no_grad():
#             sample_processed = sample if preprocess is None else preprocess(sample)
#             sample_regularized = sample if regularization is None else regularization(sample_processed)
#             y_hat = objective(sample_regularized)
#             loss = y_hat.mean()
#             log.info(f'[iter: 0/{updates}][loss: {loss.item()}]')

#     n_verbose = updates // 10

#     if threshold is not None:
#         threshold_crossed_i = np.zeros(len(sample), dtype=np.bool)
#         sample_optimized = torch.empty_like(sample)

#     for i in range(1, updates+1):
#         sample_processed = sample if preprocess is None else preprocess(sample)
#         sample_regularized = sample if regularization is None else regularization(sample_processed)
#         y_hat = objective(sample_regularized)
#         loss = y_hat.mean()

#         # Note this is threshold after preprocessing & regularization(!!)
#         if threshold is not None:
#             with torch.no_grad():
#                 # Store sample when threshold passed
#                 threshold_crossed_i_ = (y_hat.detach().cpu().numpy() < threshold)
#                 store_i = ~threshold_crossed_i&threshold_crossed_i_
#                 sample_optimized[store_i] = sample[store_i].detach()
#                 threshold_crossed_i[store_i] = True

#                 if store_i.any():
#                     log.info('threshold!')
#                     log.info(store_i, sample_optimized[store_i].detach(), sample_regularized[store_i].detach(), y_hat)
#                 # # if threshold_crossed_i_.any():
#                 # #     log.info('threshold!')
#                 # #     break

#                 if threshold_crossed_i_.all():
#                     log.info(f'Threshold crossed. [-loss: ({-loss.item()}) > threshold: ({threshold})]')
#                     break

#         optimizer_.zero_grad()
#         loss.backward()
#         optimizer_.step()
#         if i%n_verbose==0:
#             log.info(f'[iter: {i}/{updates}][loss: {loss.item()}]')

#     # Should return preprocessed sample?
#     if threshold is not None:
#         sample_optimized[~threshold_crossed_i] = sample[~threshold_crossed_i].detach()
#         return sample_optimized.to(device_original)
#     else:
#         return sample.detach().to(device_original)

# def optimize(objective, sample, updates, Optimizer, optimizer_kwargs, preprocess=None, regularization=None, threshold=None, return_best=False, verbose=False, batch_size=None):
#     """
#     Optimize sample
#     Copies the sample and returns the optimized tensor.

#     Parameters
#     ----------
#     arg :
#         dd
#     Returns
#     -------
#     sample : torch.Tensor which is optimized

#     References
#     ----------
#     .. [1] `Feature Visualization
#            <https://distill.pub/2017/feature-visualization/>`_
#     Examples
#     --------
#     >>>
#     """
#     if batch_size is None:
#         sample_optimized = _optimize(objective=objective, sample=sample, updates=updates, preprocess=preprocess, regularization=regularization, Optimizer=Optimizer, optimizer_kwargs=optimizer_kwargs, threshold=threshold, return_best=return_best, verbose=verbose)
#     else:
#         if verbose: log.info('Run in batches')
#         loader = D.DataLoader(sample, batch_size=batch_size, drop_last=False)
#         l_sample_optimized = []
#         for sample_batch in loader:
#             sample_optimized = _optimize(objective=objective, sample=sample, updates=updates, preprocess=preprocess, regularization=regularization, Optimizer=Optimizer, optimizer_kwargs=optimizer_kwargs, threshold=threshold, return_best=return_best, verbose=verbose)
#             l_sample_optimized.append(sample_optimized)
#         sample_optimized = torch.stack(l_sample_optimized, axis=0)
#     return sample_optimized



from copy import deepcopy as dcopy
from functools import partial
import inspect
import logging
import multiprocessing
import warnings

import numpy as np
import sklearn.metrics as metrics
import torch

import tools as T
import tools.torch
import tools.numpy
import torch.utils.data as D

# %%
log = logging.getLogger(__name__)

# %%
'''
Testing functions

give x and produce y_hat, y_score, y_pred, etc...
'''
def transfer_misc(model, dataset, verbose=False):
    misc_temp = T.TDict()
    transfer_log = []

    if 'get_f' in model.misc:
        transfer_log.append('get_f')
        misc_temp.get_f, dataset.get_f = dataset.get_f, model.misc.get_f

    if len(transfer_log)!=0:
        log.info(f'Transferring misc from model ({model.name}) -> dataset: {transfer_log}')

    return dataset, misc_temp

def inverse_transfer_misc(misc_temp, dataset):
    transfer_log = []
    if 'get_f' in misc_temp:
        transfer_log.append('get_f')
        dataset.get_f = misc_temp.get_f

    if len(transfer_log)!=0:
        log.info(f'Inverse transferring misc -> dataset: {transfer_log}')

    return dataset

def reproducible_worker_dict():
    '''Generate separate random number generators for workers,
    so that the global random state is not consumed,
    thereby ensuring reproducibility'''
    import random
    def seed_worker(worker_id):
        worker_seed = torch.initial_seed() % 2**32
        np.random.seed(worker_seed)
        random.seed(worker_seed)

    g = torch.Generator()
    g.manual_seed(0)
    return {'worker_init_fn': seed_worker, 'generator': g}

# DEPRECATED: use get_forward_f instead
def _test_assertion(dataset, targets_type, forward_f, result_dict, keep_x):
    # Safety check
    if forward_f is not None and result_dict is not None:
        assert callable(forward_f) and issubclass(type(result_dict), dict), f'forward_f must be callable and result_dict must be dict-like'
    elif forward_f is None and result_dict is None:
        # Infer targets_type and create corresponding forward_f and result_dict
        if targets_type is None:
            if hasattr(dataset, 'targets_type'):
                targets_type = dataset.targets_type
            else:
                raise Exception('When forward_f and result_dict is not given, targets_type must be given or the dataset must have the attribute "targets_type"')
        forward_f, result_dict = get_forward_f(targets_type, keep_x=keep_x)
    else:
        raise Exception(f'forward_f and result_dict must either both be provided or None, received [forward_f: {forward_f}][result_dict: {result_dict}]')
    return forward_f, result_dict

# def forward_network(network, dataset, forward_f=None, batch_size=64, targets_type=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
def forward_network(network, dataset, forward_f=None, batch_size=64, targets_type=None, num_workers=0, amp=False):
    """
    Forward pass of the network on the dataset.
    Takes care of device, dataloader, and mixed precision.
    
    Parameters
    ----------
    network: torch.nn.Module
        Network to be tested.
    dataset : torch.utils.data.Dataset
        Dataset to be tested.
    forward_f : callable, default=None
        forward pass function to pass the data through the network.
        If None, either inferred from dataset.targets_type or the data is just fed to the network and returned as output of the network (probably tensor/tuple not dict)

        (Arguments: network, data[, device])
        (Returns: dictionary or tuple of tensors)

    Returns
    -------
    result : ndarrays corresponding to the outputs from the forward_f 
        returns tuple or dict, based on forward_f return
    """

    # forward_f check
    if forward_f is None:
        if targets_type is None and hasattr(dataset, 'targets_type'):
            targets_type = dataset.targets_type
            forward_f = get_forward_f(targets_type)
            
        else:
            forward_f = _forward
    forward_f = partial(forward_f, network=network)

    device = T.torch.get_device(network) # Test on the network's device
    network.eval()

    # dataset, misc_temp = transfer_misc(node, dataset) # Assume the dataset is consistent
    kwargs_dataloader = reproducible_worker_dict()
    test_loader = D.DataLoader(dataset, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers, **kwargs_dataloader)

    l_results = []
    with torch.no_grad():
        # Mixed precision for acceleration
        if amp:
            with torch.autocast(device_type=device.type):
                for data in test_loader:
                    data = T.torch.to(data, device)
                    result = function_call_type_matching(f=forward_f, args=data)
                    result = T.torch.to(result, device='cpu')
                    l_results.append(result)
        else:
            for data in test_loader:
                data = T.torch.to(data, device)
                result = function_call_type_matching(f=forward_f, args=data)
                result = T.torch.to(result, device='cpu')
                l_results.append(result)

    # Formatting: Concatenate, (optional amp), numpy
    if isinstance(result, dict): 
        result = T.merge_dict(l_results)
        result = {k: torch.cat(v, dim=0) if len(v)>0 else v for k, v in result.items()}
        if amp: result = {k: v.to(torch.float16) for k, v in result}
        result = {k: v.numpy() for k, v in result.items()}
    elif isinstance(result, (list, tuple)):
        result = T.merge_tuple(l_results)
        result = tuple(torch.cat(v, dim=0) if len(v)>0 else v for v in result)
        if amp: result = tuple(v.to(torch.float16) for v in result)
        result = tuple(v.numpy() for v in result)
    elif isinstance(result, torch.Tensor): # Single tensor return
        result = torch.cat(l_results, dim=0)
        if amp: result = result.to(torch.float16)
        result = result.numpy()
    else:
        raise Exception(f'Unknown data type from forward_f return: {type(result)}')

    return result

def forward_model(model, dataset, forward_f=None, batch_size=64, targets_type=None, num_workers=0, amp=False):
    '''
    wrapper around model.
    transfers get_f (preprocessing modules) of model to dataset, and returns them to original dataset after inference.
    '''
    network = model.network
    dataset, misc_temp = transfer_misc(model, dataset)
    result = forward_network(network=network, dataset=dataset, forward_f=forward_f, batch_size=batch_size, targets_type=targets_type, num_workers=num_workers, amp=amp)
    dataset = inverse_transfer_misc(misc_temp, dataset)

    return result

# def multicall(result, functions):
def evaluate(result, metrics):
    if isinstance(metrics, dict):
        try:
            scores = {}
            for metric_name, metric in metrics.items():
                scores[metric_name] = function_call_type_matching(f=metric, args=result)
                # if isinstance(result, dict):
                #     # Only pass the arguments that are in the metric function
                #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metric).parameters.keys()}
                #     scores[metric_name] = metric(**kwargs)
                # elif isinstance(result, (list, tuple)):
                #     scores[metric_name] = metric(*result)
                # else:
                #     scores[metric_name] = metric(result)
        except Exception as e:
            if isinstance(result, dict):
                message = f'Error at metric: {metric}, result.keys(): {result.keys()}\n'
            elif isinstance(result, (list, tuple)):
                message = f'Error at metric: {metric}, len(result): {len(result)}\n'
            else:
                message = f'Error at metric: {metric}, type(result): {type(result)}\n'
            e.add_note(message)
            raise e
    elif isinstance(metrics, (list, tuple)):
        scores = []
        for metric in metrics:
            scores.append(function_call_type_matching(f=metric, args=result))
            # if isinstance(result, dict):
            #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metric).parameters.keys()}
            #     scores.append(metric(**kwargs))
            # elif isinstance(result, (list, tuple)):
            #     scores.append(metric(*result))
            # else:
            #     scores.append(metric(result))
    else: # Single metric
        scores = function_call_type_matching(f=metrics, args=result)
        # if isinstance(result, dict):
        #     kwargs = {k:v for k, v in result.items() if k in inspect.signature(metrics).parameters.keys()}
        #     if len(kwargs) < len(result):
        #         warnings.warn('Some arguments in result are not used in the metric function, when only single metric was passed')
        #     scores = metrics(**kwargs)
        # elif isinstance(result, (list, tuple)):
        #     scores = metrics(*result)
        # else:
        #     scores = metrics(result)
    return scores

targets_type_list = [None, 'categorical', 'multihead_classification', 'autoencode', 'regression'] # move this to utils?
def get_forward_f(targets_type):
    assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
    if targets_type is None or targets_type == 'regression':
        forward_f = _forward_base
        result_keys = ['y_true', 'y_pred']
    elif targets_type == 'categorical':
        forward_f = _forward_categorical
        result_keys = ['y_true', 'y_hat', 'y_score', 'y_pred'] # y_hat vs y_logit?
    elif targets_type == 'multihead_classification':
        forward_f = _forward_multihead_categorical
        result_keys = ['y_true', 'y_hat', 'y_score', 'y_pred']
    elif targets_type == 'autoencode':
        forward_f = _forward_autoencode
        result_keys = ['x', 'z', 'x_hat']
    else:
        raise Exception(f'unknown targets_type: {targets_type}')
    
    log.info(f'get_forward_f: targets_type: {targets_type}, keys in result (dict): {result_keys}')

    return forward_f

def function_call_type_matching(f, args):
    try:
        if isinstance(args, dict):
            function_arguments = inspect.signature(f).parameters.keys()
            if 'kwargs' not in function_arguments: # Will raise error if undefined arguments are passed
                # Only pass the arguments that are in the function
                kwargs = {k:v for k, v in args.items() if k in function_arguments}
            else: # kwargs will allow undefined arguments of f
                kwargs = args
            return f(**kwargs)
        elif isinstance(args, (list, tuple)):
            return f(*args)
        else:
            return f(args)
    except Exception as e:
        function_arguments = inspect.signature(f).parameters.keys()
        if isinstance(args, dict):
            args = args.keys()
        elif isinstance(args, (list, tuple)):
            args = f'tuple of len({len(args)})'
        else:
            args = args

        print(f'f: {f}, f_args: {function_arguments}, args: {args}')
        # e.add_note()
        raise e

def _forward(network, data):
    return network(data)
    # if isinstance(data, dict):
    #     return network(**data)
    # elif isinstance(data, (list, tuple)):
    #     return network(*data)
    # else:
    #     return network(data)

def _forward_base(network, data):
    # device = device if device is not None else T.torch.get_device(network)
    x = data['x']
    y = data['y']
    y_pred = network(x)

    result_dict = {
        'y_true': y,
        'y_pred': y_pred
    }
    # if keep_x: result_dict['x'] = x
    return result_dict

def _forward_categorical(network, data):
    # Is y_hat necessary? for torch_crossentropyloss? any other methods?
    x = data['x']
    y = data['y']
    y_hat = network(x)
    y_score = torch.softmax(y_hat, dim=1) # (N, n_classes)

    result_dict = {
        'y_true': y,
        'y_hat': y_hat,
        'y_score': y_score,
        'y_pred': y_score.argmax(axis=1)
    }
    # if keep_x: result_dict['x'] = x
    return result_dict

def _forward_multihead_categorical(network, data):
    x = data['x']
    y = data['y']
    y_hat = network(x)
    y_score = torch.softmax(y_hat, dim=-1) # (N, subtype, n_classes)

    result_dict = {
        'y_true': y,
        'y_hat': y_hat,
        'y_score': y_score,
        'y_pred': y_score.argmax(axis=-1)
    }
    # if keep_x: result_dict['x'] = x
    return result_dict

def _forward_autoencode(network, data):
    x = data['x']
    z = network.encoder(x)
    x_hat = torch.sigmoid(network.decoder(z))

    result_dict = {
        'x': x,
        'z': z,
        'x_hat': x_hat
    }
    return result_dict

# %%
'''
Scorers

They all receive dict of numpy arrays.
Predefined keys are:
'y_true', 'y_hat', 'y_pred'

Example
-------
result = {'y_true': y_true, 'y_pred': y_pred}
classification_score(result)

scores = {'scorer type': score, ...}


# order of arguments follow sklearn convention: y_true, y_hat/pred

Note
----
pytorch metrics receive (y_hat/pred, y_true), whereas sklearn receive (y_true, y_hat/pred)
'''

# Regresison scorers
def l1_score(result): # alias
    return metrics.mean_absolute_error(result['y_true'], result['y_hat'])

def l2_score(result): # alias
    return metrics.mean_squared_error(result['y_true'], result['y_hat'])

def mse_score(result): # alias
    return metrics.mean_squared_error(result['y_true'], result['y_hat'])

def r2_score(result):
    return metrics.r2_score(result['y_true'], result['y_hat'])

def p_norm_score(result, p=2): # When using this with Validation object, use functools.partial() to fix ``p``
    score = ((np.abs(result['y_hat']-result['y_true'])**p).sum())**(1/p)
    return score

# Classification score
def accuracy_score(result):
    return metrics.accuracy_score(result['y_true'], result['y_pred'])

# def accuracy_score_y_hat(result): # Use this for Validation object, which feeds y_hat to all of its scorers.
#     return metrics.accuracy_score(result['y_true'], result['y_hat'].argmax(axis=1))

# Binary classification
def sensitivity_score(result): # alias
    '''sensitivity == recall == tpr'''
    return metrics.recall_score(result['y_true'], result['y_pred'])

def specificity_score(result):
    (tn, fp), (fn, tp) = metrics.confusion_matrix(result['y_true'], result['y_pred'])
    if (tn+fp)==0:
        warnings.warn('invalid value in specificity_score, setting to 0.0')
        return 0
    return tn / (tn+fp)

# Multiclass classification
def classification_report_full(result, ovr=True):
    '''
    Adds additional metrics to sklearn.classification_report
    '''
    y_score = result['y_score'] if 'y_score' in result else None
    y_true, y_pred = result['y_true'], result['y_pred']
    
    scores = metrics.classification_report(y_true, y_pred, output_dict=True)
    scores['mcc'] = metrics.matthews_corrcoef(y_true, y_pred)

    if ovr:
        scores_ovr = {k:dcopy(v) for k, v in scores.items() if k.isnumeric()}
        scores_all = {k:dcopy(v) for k, v in scores.items() if not k.isnumeric()}

        more_scorers_y_pred = {'sensitivity': sensitivity_score, 'specificity': specificity_score, 'accuracy': accuracy_score} # Optimize to reduce redundant computations?
        more_scorers_y_score = {}

        # Additional metrics
        for c in scores_ovr.keys():
            c_int = int(c)
            y_true__c, y_pred_c = y_true==c_int, y_pred==c_int
            for scorer_name, scorer in more_scorers_y_pred.items():
                scores_ovr[c][scorer_name] = scorer({'y_true': y_true__c, 'y_pred': y_pred_c})

        if y_score is not None:
            more_scorers_y_score['auroc'] = metrics.roc_auc_score
            for c in scores_ovr.keys():
                c_int = int(c)
                y_true_ = y_true==c_int
                y_score_ = y_score[:, c_int]
                for scorer_name, scorer in more_scorers_y_score.items():
                    scores_ovr[c][scorer_name] = scorer(y_true_, y_score_)

        # summary
        more_scorers = list(more_scorers_y_pred.keys()) + list(more_scorers_y_score.keys())
        for scorer_name in more_scorers:
            scores_all['macro avg'][scorer_name] = np.mean([scores_ovr_[scorer_name] for scores_ovr_ in scores_ovr.values()])
            scores_all['weighted avg'][scorer_name] = np.sum([scores_ovr_[scorer_name]*scores_ovr_['support'] for scores_ovr_ in scores_ovr.values()]) / scores_all['weighted avg']['support']

        scores.update(scores_ovr)
        scores.update(scores_all)

    return scores

# Multihead classification score
def multihead_accuracy_score(y_pred, y):
    '''
    :param y_pred: array of shape (N, subtype)
    :param y: array of shape (N, subtype)
    '''
    assert y_pred.ndim==2 and y.ndim==2
    score = [accuracy_score(y_pred_, y_) for y_pred_, y_ in zip(y_pred.T, y.T)]
    score = np.mean(score)
    return score

def multihead_accuracy_score_score(y_score, y):
    '''
    :param y_pred: array of shape (N, subtype, n_classes)
    :param y: array of shape (N, subtype)
    '''
    assert y_score.ndim==3 and y.ndim==2
    y_pred = y_score.argmax(axis=-1)
    return multihead_accuracy_score(y_pred, y)

# %%
# All scores for each task type
def regression_score(result):
    '''
    :param result: dict with the following keys: [y_hat, y]
    '''
    y_hat, y = result['y_hat'], result['y']
    l1 = l1_score(y_hat, y)
    mse = mse_score(y_hat, y)
    r2 = r2_score(y_hat, y)

    scores = {
    'l1': l1,
    'mse': mse,
    'r2': r2
    }

    return scores

def classification_score(result, multi_class='ovr', discard_ovr=False):
    '''
    :param result: dict with the following keys: [y_pred, y_score, y]


    y_score: array of shape (N, n_classes)
    y: array of shape (N,)
    '''
    y_pred, y_score, y_true = result['y_pred'], result['y_score'], result['y']

    # classification_result = classification_report_full(y_true, y_pred, discard_ovr=discard_ovr)
    classification_result = classification_report_full(result, discard_ovr=discard_ovr)
    auroc = metrics.roc_auc_score(y_true, y_score, multi_class=multi_class)
    kappa = metrics.cohen_kappa_score(y_true, y_pred)
    m_cc = metrics.matthews_corrcoef(y_true, y_pred)
    c_matrix = metrics.confusion_matrix(y_true, y_pred)

    scores = {
    'auroc': auroc,
    'kappa': kappa,
    'm_cc': m_cc,
    'c_matrix': c_matrix,
    }
    classification_result.update(scores)
    return scores

def multihead_classification_score(result):
    '''
    :param result: dict with the following keys: [y_pred, y_score, y]


    y_score: array of shape (N, subtype, n_classes)
    y: array of shape (N, subtype)
    '''
    y_pred, y_score, y = result['y_pred'], result['y_score'], result['y']
    multihead_accuracy = multihead_accuracy_score(y_pred, y)

    scores = {
    'multihead_accuracy_score': multihead_accuracy
    }
    return scores


# Wrapper function for hydra.utils.instantiate
class Score_wrapper:
    '''
    wraps around scorers to:
    1. instantiate an object using hydra.utils.instantiate()
    2. adjust scorer functions such that raw input can be delivered to scorer functions (y_pred or y_score can be delivered to any type of scorers)
    '''
    f_dict = {
    # regression
    'l1_score': l1_score,
    'mse_score': mse_score,
    'r2_score': r2_score,

    # multihead classification
    'multihead_accuracy_score': multihead_accuracy_score_score,
    }
    def __init__(self, func):
        self.func = self.f_dict[func]

    def __call__(self, y_hat, y):
        return self.func(y_hat, y)
# %%
'''
May be deprecated
'''
def save_score():
    f

def _pr_score(y_true, y_score, threshold):
    y_pred = T.numpy.binarize(y_score, threshold)
    pr = metrics.precision_score(y_true, y_pred)
    rec = metrics.recall_score(y_true, y_pred)
    return pr, rec

def precision_recall_curve_all(y_true, y_score, num_workers=None):
    '''
    compute precision-recall curve for all threshold.
    Because sklearn.metrics.precision_recall_curve does not return full set of thresholds
    sklearn drops the results after when recall hits 1.

    -> is this function really necessary?
    -> may be deprecated
    '''
    thresholds = np.sort(np.unique(y_score)) # increasing threshold
    thresholds = np.append(thresholds,1) # Since tnp.binarize binarizes with y_pred[y_score>=threshold] == 1
    if num_workers==0:
        prs, recs = [], []
        for t in thresholds:
            y_pred = T.numpy.binarize(y_score, threshold)
            prs.append(metrics.precision_score(y_true, y_pred))
            recs.append(metrics.recall_score(y_true, y_pred))
    elif num_workers is None or T.isint(num_workers):
        with multiprocessing.Pool() as p:
            l_pr = p.starmap(_pr_score, zip(it.repeat(y_true, len(thresholds)), it.repeat(y_score, len(thresholds)), thresholds))

    return prs, recs, thresholds
